import pytest
import numpy as np
import astropy.units as u  # type: ignore[import]
import named_arrays as na  # type: ignore[import]
import optika


class AbstractTestAbstractTransform:
    def test_matrix(self, transform: optika.transforms.AbstractTransform):
        if transform.matrix is not None:
            assert isinstance(
                transform.matrix,
                na.AbstractCartesian3dMatrixArray,
            )
            assert isinstance(
                transform.matrix.to(u.dimensionless_unscaled),
                na.AbstractCartesian3dMatrixArray,
            )

    def test_vector(self, transform: optika.transforms.AbstractTransform):
        if transform.vector is not None:
            assert isinstance(
                transform.vector,
                na.AbstractCartesian3dVectorArray,
            )
            assert isinstance(
                transform.vector.to(u.mm),
                na.AbstractCartesian3dVectorArray,
            )

    def test__call__(self, transform: optika.transforms.AbstractTransform):
        x = na.Cartesian3dVectorArray(x=1, y=-2, z=3) * u.m
        y = transform(x)
        assert isinstance(y, na.AbstractCartesian3dVectorArray)

    def test_inverse(self, transform: optika.transforms.AbstractTransform):
        x = na.Cartesian3dVectorArray(x=1, y=-2, z=3) * u.m
        y = transform(x)
        z = transform.inverse(y)
        assert np.allclose(x, z)

    @pytest.mark.parametrize(
        argnames="transform_2",
        argvalues=[
            optika.transforms.Translation(na.Cartesian3dVectorArray(5) * u.mm),
            optika.transforms.RotationX(53 * u.deg),
            optika.transforms.TransformList(
                [
                    optika.transforms.Translation(na.Cartesian3dVectorArray(5) * u.mm),
                    optika.transforms.RotationX(53 * u.deg),
                ]
            ),
        ],
    )
    class TestMatmul:
        def test__matmul__(
            self,
            transform: optika.transforms.AbstractTransform,
            transform_2: optika.transforms.AbstractTransform,
        ):
            result = transform @ transform_2
            assert isinstance(result, optika.transforms.TransformList)
            for t in result:
                assert isinstance(t, optika.transforms.AbstractTransform)

            x = na.Cartesian3dVectorUniformRandomSample(
                start=-5 * u.mm,
                stop=5 * u.mm,
                shape_random=dict(x=5, y=6),
            )
            assert np.allclose(result(x), transform(transform_2(x)))

        def test__matmul__reversed(
            self,
            transform: optika.transforms.AbstractTransform,
            transform_2: optika.transforms.AbstractTransform,
        ):
            return self.test__matmul__(
                transform=transform_2,
                transform_2=transform,
            )


displacement_parameterization = [
    na.Cartesian3dVectorArray() * u.mm,
    na.Cartesian3dVectorArray(1, -2, 3) * u.mm,
    na.Cartesian3dVectorArray(
        x=1 * u.mm,
        y=na.linspace(-5, 5, axis="translation", num=4) * u.mm,
        z=0 * u.mm,
    ),
    na.Cartesian3dVectorArray(
        x=0 * u.mm,
        y=na.linspace(-5, 5, axis="translation", num=4) * u.mm,
        z=na.NormalUncertainScalarArray(6 * u.mm, width=0.1 * u.mm),
    ),
]


@pytest.mark.parametrize(
    argnames="transform",
    argvalues=[
        optika.transforms.Translation(displacement)
        for displacement in displacement_parameterization
    ],
)
class TestTranslation(
    AbstractTestAbstractTransform,
):
    pass


angle_parameterization = [
    0 * u.deg,
    45 * u.deg,
    na.linspace(0 * u.deg, 360 * u.deg, axis="angle", num=3),
    na.NormalUncertainScalarArray(45 * u.deg, width=5 * u.deg),
]


class AbstractTestAbstractRotation(
    AbstractTestAbstractTransform,
):
    pass


@pytest.mark.parametrize(
    argnames="transform",
    argvalues=[optika.transforms.RotationX(angle) for angle in angle_parameterization],
)
class TestRotationX(
    AbstractTestAbstractRotation,
):
    pass


@pytest.mark.parametrize(
    argnames="transform",
    argvalues=[optika.transforms.RotationY(angle) for angle in angle_parameterization],
)
class TestRotationY(
    AbstractTestAbstractRotation,
):
    pass


@pytest.mark.parametrize(
    argnames="transform",
    argvalues=[optika.transforms.RotationZ(angle) for angle in angle_parameterization],
)
class TestRotationZ(
    AbstractTestAbstractRotation,
):
    pass


@pytest.mark.parametrize(
    argnames="transform",
    argvalues=[
        optika.transforms.TransformList(
            [
                optika.transforms.Translation(displacement),
                optika.transforms.RotationZ(angle),
                optika.transforms.Translation(-displacement),
                optika.transforms.RotationY(-angle),
            ]
        )
        for displacement in displacement_parameterization
        for angle in angle_parameterization
    ],
)
class TestTransformList(
    AbstractTestAbstractTransform,
):
    def test__call__(self, transform: optika.transforms.TransformList):  # type: ignore[override]
        super().test__call__(transform=transform)
        x = na.Cartesian3dVectorArray() * u.m
        b = transform(x)
        c = x
        for t in transform.transforms:
            c = t(c)
        assert np.allclose(b, c)


transform_parameterization = [
    None,
    optika.transforms.TransformList(
        [
            optika.transforms.Translation(na.Cartesian3dVectorArray(5) * u.mm),
            optika.transforms.RotationZ(53 * u.deg),
            optika.transforms.Translation(na.Cartesian3dVectorArray(5) * u.mm),
        ]
    ),
    optika.transforms.RotationZ(
        na.ScalarLinearSpace(0 * u.deg, 90 * u.deg, axis="transform", num=3)
    ),
    optika.transforms.RotationZ(
        na.NormalUncertainScalarArray(53 * u.deg, width=5 * u.deg)
    ),
]


class AbstractTestTransformable:
    def test_transform(self, a: optika.transforms.Transformable):
        if a.transform is not None:
            assert isinstance(a.transform, optika.transforms.AbstractTransform)



from __future__ import annotations
from typing import Generic, TypeVar, Iterator, Iterable, Any, overload
from typing_extensions import Self
import copy
import dataclasses
import numpy as np
import astropy.units as u
import named_arrays as na

__all__ = [
    "Printable",
    "DataclassList",
]

ItemT = TypeVar("ItemT")


@dataclasses.dataclass(repr=False)
class Printable:
    @classmethod
    def _val_to_string(
        cls,
        val: Any,
        pre: str,
        tab: str,
        field_str: str,
    ) -> str:
        if isinstance(val, Printable):
            val_str = val.to_string(prefix=f"{pre}{tab}")
        elif isinstance(val, na.AbstractArray):
            val_str = val.to_string(prefix=f"{pre}{tab}")
        elif isinstance(val, np.ndarray):
            val_str = np.array2string(
                a=val,
                separator=", ",
                prefix=field_str,
            )
            if isinstance(val, u.Quantity):
                val_str = f"{val_str} {val.unit}"
        elif isinstance(val, list):
            val_str = f"[\n"
            for v in val:
                val_str += f"{pre}{tab}{tab}"
                val_str += cls._val_to_string(
                    val=v,
                    pre=f"{pre}{tab}",
                    tab=tab,
                    field_str=f"{pre}{tab}",
                )
                val_str += ",\n"
            val_str += f"{pre}{tab}]"
        else:
            val_str = repr(val)

        return val_str

    def to_string(
        self,
        prefix: None | str = None,
    ) -> str:
        """
        Public-facing version of the ``__repr__`` method that allows for
        defining a prefix string, which can be used to calculate how much
        whitespace to add to the beginning of each line of the result.

        Parameters
        ----------
        prefix
            an optional string, the length of which is used to calculate how
            much whitespace to add to the result.
        """

        fields = dataclasses.fields(self)

        delim_field = "\n"
        pre = " " * len(prefix) if prefix is not None else ""
        tab = " " * 4

        result = f"{self.__class__.__qualname__}(\n"

        for i, f in enumerate(fields):
            field_str = f"{pre}{tab}{f.name}="
            val = getattr(self, f.name)
            val_str = self._val_to_string(
                val=val,
                pre=pre,
                tab=tab,
                field_str=field_str,
            )
            field_str += val_str
            field_str += f",{delim_field}"
            result += field_str
        result += f"{pre})"
        return result

    def __repr__(self):
        return self.to_string()


@dataclasses.dataclass
class DataclassList(
    Generic[ItemT],
):
    data: list[ItemT] = dataclasses.field(default_factory=list)

    def __contains__(self, item: ItemT) -> bool:
        return self.data.__contains__(item)

    def __iter__(self) -> Iterator[ItemT]:
        return self.data.__iter__()

    def __reversed__(self) -> Iterator[ItemT]:
        return self.data.__reversed__()

    @overload
    def __getitem__(self, item: int) -> ItemT:
        ...

    @overload
    def __getitem__(self: Self, item: slice) -> Self:
        ...

    def __getitem__(self: Self, item: int | slice) -> ItemT | Self:
        if isinstance(item, slice):
            other = copy.copy(self)
            other.data = self.data.__getitem__(item)
            return other
        else:
            return self.data.__getitem__(item)

    @overload
    def __setitem__(self, key: int, value: ItemT):
        ...

    @overload
    def __setitem__(self, key: slice, value: Iterable[ItemT]):
        ...

    def __setitem__(self, key, value):
        self.data.__setitem__(key, value)

    def __delitem__(self, key: int | slice):
        self.data.__delitem__(key)

    def __len__(self) -> int:
        return self.data.__len__()

    def __add__(self, other: DataclassList):
        new_self = copy.copy(self)
        new_self.data = self.data.__add__(other.data)
        return new_self

    def index(self, value: ItemT) -> int:
        return self.data.index(value)

    def count(self, value: ItemT) -> int:
        return self.data.count(value)

    def append(self, item: ItemT) -> None:
        self.data.append(item)

    def reverse(self):
        self.data.reverse()



import abc
import dataclasses
import numpy as np
import numpy.typing as npt
import matplotlib.axes
import matplotlib.lines
import matplotlib.pyplot as plt
import astropy.units as u
import named_arrays as na
import optika.mixins
import optika.plotting

__all__ = [
    "AbstractAperture",
    "CircularAperture",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractAperture(
    optika.mixins.Printable,
    optika.plotting.Plottable,
    optika.transforms.Transformable,
):
    @property
    @abc.abstractmethod
    def samples_per_side(self):
        """
        number of samples per side of the polygon representing the aperture
        """

    @property
    @abc.abstractmethod
    def active(self):
        """
        flag controlling whether the aperture can clip rays during a raytrace
        """

    @property
    @abc.abstractmethod
    def inverted(self):
        """
        flag controlling whether the interior or the exterior of the aperture
        allows light to pass through.
        If :obj:`True`, the interior of the aperture allows light to pass
        through.
        If :obj:`False`, the exterior of the aperture allows light to pass
        through.
        """

    @abc.abstractmethod
    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        """
        Check if a given point is inside the aperture.

        Parameters
        ----------
        position
            the points to check
        """

    def clip_rays(self, rays: optika.rays.RayVectorArray):
        mask = self(rays.position)
        rays = rays.copy_shallow()
        rays.intensity = rays.intensity * mask
        return rays

    @property
    @abc.abstractmethod
    def bound_lower(self) -> na.AbstractCartesian3dVectorArray:
        """
        The lower-left corner of the aperture's rectangular footprint
        """

    @property
    @abc.abstractmethod
    def bound_upper(self) -> na.AbstractCartesian3dVectorArray:
        """
        The upper-right corner of the aperture's rectangular footprint
        """

    @property
    @abc.abstractmethod
    def vertices(self) -> None | na.AbstractCartesian3dVectorArray:
        """
        The vertices of the polygon representing this aperture
        """

    @property
    @abc.abstractmethod
    def wire(self) -> na.AbstractCartesian3dVectorArray:
        """
        The sequence of points representing this aperture
        """

    def plot(
        self,
        ax: None | matplotlib.axes.Axes | na.ScalarArray[npt.NDArray] = None,
        transform: None | optika.transforms.AbstractTransform = None,
        component_map: dict[str, str] = None,
        sag: None | optika.sags.AbstractSag = None,
        **kwargs,
    ) -> None | na.ScalarArray[npt.NDArray[None | matplotlib.lines.Line2D]]:
        if ax is None:
            ax = plt.gca()
        ax = na.as_named_array(ax)

        if component_map is None:
            component_map = dict()
        component_map = dict(x="x", y="y", z="z") | component_map

        wire = self.wire.explicit

        if not wire.length.unit_normalized.is_equivalent(u.mm):
            return None

        if sag is not None:
            wire.z = sag(wire)

        if transform is not None:
            wire = transform(wire)

        kwargs_plot = self.kwargs_plot
        if kwargs_plot is None:
            kwargs_plot = dict()

        kwargs = kwargs_plot | kwargs

        wire = wire.components
        wire = na.Cartesian3dVectorArray(
            x=wire[component_map["x"]],
            y=wire[component_map["y"]],
            z=wire[component_map["z"]],
        )

        return na.plt.plot(
            wire,
            ax=ax,
            axis="wire",
            **kwargs,
        )


@dataclasses.dataclass(eq=False, repr=False)
class CircularAperture(
    AbstractAperture,
):
    """
    A circular aperture or obscuration

    Examples
    --------

    Plot a single circular aperture

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import numpy as np
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        aperture = optika.apertures.CircularAperture(50 * u.mm)

        with astropy.visualization.quantity_support():
            plt.figure()
            plt.gca().set_aspect("equal")
            aperture.plot(color="black")

    |

    Plot an array of circular apertures, similar to the configuration of the
    `Giant Magellan Telescope <https://en.wikipedia.org/wiki/Giant_Magellan_Telescope>`_.

    .. jupyter-execute::

        diameter = 8.417 * u.m
        radius = diameter / 2

        angle = na.linspace(0, 360, axis="segment", num=6, endpoint=False) * u.deg

        displacement = na.Cartesian3dVectorArray(
            x=diameter * np.cos(angle),
            y=diameter * np.sin(angle),
        )
        displacement = np.concatenate([
            na.Cartesian3dVectorArray() * u.mm,
            displacement
        ], axis="segment")

        aperture = optika.apertures.CircularAperture(
            radius=radius,
            transform=optika.transforms.Translation(displacement),
        )

        with astropy.visualization.quantity_support():
            plt.figure()
            plt.gca().set_aspect("equal")
            aperture.plot(color="black")
    """

    radius: u.Quantity | na.AbstractScalar = 0 * u.mm
    """the radius of the aperture"""

    samples_per_side: int = 101
    active: bool | na.AbstractScalar = True
    inverted: bool | na.AbstractScalar = False
    transform: None | optika.transforms.AbstractTransform = None
    kwargs_plot: None | dict = None

    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        radius = self.radius
        active = self.active
        inverted = self.inverted
        if self.transform is not None:
            position = self.transform.inverse(position)

        shape = na.shape_broadcasted(radius, active, inverted, position)

        radius = na.broadcast_to(radius, shape)
        active = na.broadcast_to(active, shape)
        inverted = na.broadcast_to(inverted, shape)
        position = na.broadcast_to(position, shape)

        mask = position.length <= radius

        mask[inverted] = ~mask[inverted]
        mask[~active] = True

        return mask

    @property
    def bound_lower(self) -> na.Cartesian3dVectorArray:
        result = na.Cartesian3dVectorArray(
            x=-self.radius,
            y=-self.radius,
            z=0 * self.radius.unit,
        )
        if self.transform is not None:
            result = self.transform(result, use_matrix=False)
        return result

    @property
    def bound_upper(self) -> na.Cartesian3dVectorArray:
        result = na.Cartesian3dVectorArray(
            x=self.radius,
            y=self.radius,
            z=0 * self.radius.unit,
        )
        if self.transform is not None:
            result = self.transform(result, use_matrix=False)
        return result

    @property
    def vertices(self) -> None:
        return None

    @property
    def wire(self) -> na.Cartesian3dVectorArray:
        az = na.linspace(
            start=0 * u.deg,
            stop=360 * u.deg,
            axis="wire",
            num=self.samples_per_side,
        )
        result = na.Cartesian3dVectorArray(
            x=self.radius * np.cos(az),
            y=self.radius * np.sin(az),
            z=0 * self.radius.unit,
        )
        if self.transform is not None:
            result = self.transform(result)
        return result


@dataclasses.dataclass(eq=False, repr=False)
class AbstractPolygonalAperture(
    AbstractAperture,
):
    """
    Base class for any type of polygonal aperture
    """

    @property
    def bound_lower(self) -> na.AbstractCartesian3dVectorArray:
        return self.vertices.min(axis="vertex")

    @property
    def bound_upper(self) -> na.AbstractCartesian3dVectorArray:
        return self.vertices.max(axis="vertex")

    @property
    def wire(self) -> na.Cartesian3dVectorArray:
        vertices = self.vertices.broadcasted
        num_vertices = vertices.shape["vertex"]
        ind_rolled = dict(vertex=na.arange(0, num_vertices, axis="vertex") - 1)
        vertices_left = vertices[ind_rolled]
        vertices_right = vertices
        diff = vertices_right - vertices_left
        t = na.linspace(
            start=0,
            stop=1,
            axis="wire",
            num=self.samples_per_side,
            endpoint=False,
        )
        wire = vertices_left + diff * t
        wire = wire.combine_axes(axes=("vertex", "wire"), axis_new="wire")
        return wire


@dataclasses.dataclass(eq=False, repr=False)
class RectangularAperture(
    AbstractPolygonalAperture,
):
    """
    A rectangular aperture or obscuration

    Examples
    --------

    Create a square aperture by setting :attr:`half_width` to a scalar value.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        half_width = 50 * u.mm

        aperture = optika.apertures.RectangularAperture(half_width)

        with astropy.visualization.quantity_support():
            plt.figure()
            plt.gca().set_aspect("equal")
            aperture.plot(color="black")

    |

    Create a rectangular aperture by setting :attr:`half_width` to an
    instance of :class:`named_arrays.AbstractCartesian2dVectorArray`.

    .. jupyter-execute::

        half_width = na.Cartesian2dVectorArray(100, 50) * u.mm

        aperture = optika.apertures.RectangularAperture(half_width)

        with astropy.visualization.quantity_support():
            plt.figure()
            plt.gca().set_aspect("equal")
            aperture.plot(color="black")

    |

    Create a grid of rectangular apertures using the :attr:`transform`
    parameter and the :class:`optika.transforms.Translation` transformation.

    .. jupyter-execute::

        pitch = 2 * half_width + 10 * u.mm

        displacement = na.Cartesian3dVectorArray(
            x=pitch.x * na.arange(0, 3, axis="aperture_x"),
            y=pitch.y * na.arange(0, 2, axis="aperture_y"),
        )

        aperture = optika.apertures.RectangularAperture(
            half_width=half_width,
            transform=optika.transforms.Translation(displacement),
        )

        with astropy.visualization.quantity_support():
            plt.figure()
            plt.gca().set_aspect("equal")
            aperture.plot(color="black")
    """

    half_width: u.Quantity | na.AbstractScalar | na.Cartesian2dVectorArray = 0 * u.mm
    """distance from the origin to a perpendicular edge"""

    samples_per_side: int = 101
    active: bool | na.AbstractScalar = True
    inverted: bool | na.AbstractScalar = False
    transform: None | optika.transforms.AbstractTransform = None
    kwargs_plot: None | dict = None

    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        bound_lower = self.bound_lower
        bound_upper = self.bound_upper
        active = self.active
        inverted = self.inverted
        if self.transform is not None:
            position = self.transform.inverse(position)

        shape = na.shape_broadcasted(
            bound_lower, bound_upper, active, inverted, position
        )

        bound_lower = na.broadcast_to(bound_lower, shape)
        bound_upper = na.broadcast_to(bound_upper, shape)
        active = na.broadcast_to(active, shape)
        inverted = na.broadcast_to(inverted, shape)
        position = na.broadcast_to(position, shape)

        mask = (bound_lower <= position) & (position <= bound_upper)
        mask = mask.x & mask.y

        mask[inverted] = ~mask[inverted]
        mask[~active] = True

        return mask

    @property
    def vertices(self):
        half_width = na.asanyarray(self.half_width, like=na.Cartesian2dVectorArray())
        r = np.sqrt(2)
        az = na.linspace(0, 360, axis="vertex", num=4, endpoint=False) * u.deg
        az = az + 45 * u.deg
        result = na.Cartesian3dVectorArray(
            x=r * np.cos(az),
            y=r * np.sin(az),
            z=0,
        )
        result.x = result.x * half_width.x
        result.y = result.y * half_width.y
        result.z = 0 * half_width.length.unit
        if self.transform is not None:
            result = self.transform(result)
        return result



from __future__ import annotations
from typing import TypeVar, Generic
import abc
import dataclasses
import named_arrays as na

__all__ = [
    "AbstractRayVectorArray",
    "RayVectorArray",
    "AbstractImplicitRayVectorArray",
]

IntensityT = TypeVar("IntensityT", bound=na.ScalarLike)
AttenuationT = TypeVar("AttenuationT", bound=na.ScalarLike)
PositionT = TypeVar("PositionT", bound=na.Cartesian3dVectorArray)
DirectionT = TypeVar("DirectionT", bound=na.Cartesian3dVectorArray)
WavelengthT = TypeVar("WavelengthT", bound=na.ScalarLike)
IndexRefractionT = TypeVar("IndexRefractionT", bound=na.ScalarLike)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractRayVectorArray(
    na.AbstractSpectralPositionalVectorArray,
):
    @property
    @abc.abstractmethod
    def direction(self) -> na.AbstractCartesian3dVectorArray:
        """
        the propagation direction of the ray expressed in
        `direction cosines <https://en.wikipedia.org/wiki/Direction_cosine>`_
        """

    @property
    @abc.abstractmethod
    def intensity(self) -> na.AbstractScalar:
        """
        the radiometric contribution of the ray
        """

    @property
    @abc.abstractmethod
    def attenuation(self) -> na.AbstractScalar:
        """
        the current
        `attenuation coefficient <https://en.wikipedia.org/wiki/Absorption_coefficient>`_
        of the medium that the ray is propagating in
        """

    @property
    @abc.abstractmethod
    def index_refraction(self):
        """
        the current index of refraction of the medium that the ray is
        traveling in
        """

    @property
    def type_abstract(self) -> type[na.AbstractArray]:
        return AbstractRayVectorArray

    @property
    def type_explicit(self) -> type[na.AbstractExplicitArray]:
        return RayVectorArray

    @property
    def type_matrix(self) -> type[na.AbstractMatrixArray]:
        raise NotImplementedError

    @property
    def explicit(self) -> RayVectorArray:
        return super().explicit


@dataclasses.dataclass(eq=False, repr=False)
class RayVectorArray(
    AbstractRayVectorArray,
    na.SpectralPositionalVectorArray[WavelengthT, PositionT],
    Generic[IntensityT, PositionT, DirectionT, WavelengthT, IndexRefractionT],
):
    direction: DirectionT = 0
    intensity: IntensityT = 0
    attenuation: AttenuationT = 0
    index_refraction: IndexRefractionT = 1

    @classmethod
    def from_scalar(
        cls,
        scalar: na.AbstractScalar,
        like: None | na.AbstractExplicitVectorArray = None,
    ) -> RayVectorArray:
        return cls(
            wavelength=scalar,
            position=scalar,
            direction=scalar,
            intensity=scalar,
            attenuation=scalar,
            index_refraction=scalar,
        )


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitRayVectorArray(
    AbstractRayVectorArray,
    na.AbstractImplicitCartesianVectorArray,
):
    @property
    def direction(self) -> na.AbstractCartesian3dVectorArray:
        return self.explicit.direction

    @property
    def intensity(self) -> na.ScalarLike:
        return self.explicit.intensity

    @property
    def attenuation(self) -> na.ScalarLike:
        return self.explicit.attenuation

    @property
    def index_refraction(self) -> na.ScalarLike:
        return self.explicit.index_refraction



from __future__ import annotations
from typing import Iterator
from typing_extensions import Self
import abc
import dataclasses
import copy
import astropy.units as u  # type: ignore[import]
import named_arrays as na  # type: ignore[import]
import optika.mixins

__all__ = [
    "AbstractTransform",
    "Translation",
    "AbstractRotation",
    "RotationX",
    "RotationY",
    "RotationZ",
    "TransformList",
    "Transformable",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractTransform(
    optika.mixins.Printable,
):
    """
    An interface for an arbitrary vector transform
    """

    @property
    @abc.abstractmethod
    def explicit(self) -> Transform:
        """
        resolve this transform into its explicit form
        """

    @property
    def matrix(self) -> na.AbstractCartesian3dMatrixArray:
        """
        The matrix component of the transformation.

        This can contribute to scaling, shearing, or rotating an instance of
        :class:`named_arrays.AbstractVectorArray`
        """
        return self.explicit.matrix

    @property
    def vector(self) -> na.AbstractCartesian3dVectorArray:
        """
        The vector component of the transformation

        This can contribute to translation of an instance of
        :class:`named_arrays.AbstractVectorArray`.
        """
        return self.explicit.vector

    def __call__(
        self,
        value: na.AbstractCartesian3dVectorArray,
        use_matrix: bool = True,
        use_vector: bool = True,
    ) -> na.AbstractCartesian3dVectorArray:
        """
        evaluate the transformation

        Parameters
        ----------
        value
            the vector to transform
        use_matrix
            a flag controlling whether the matrix component of the transformation
            is applied
        use_vector
            a flag controlling whether the vector component of the transformation
            is applied
        """

        explicit = self.explicit
        if use_matrix:
            matrix = explicit.matrix
            if matrix is not None:
                value = matrix @ value
        if use_vector:
            vector = explicit.vector
            if vector is not None:
                value = value + vector
        return value

    @abc.abstractmethod
    def __invert__(self: Self) -> Self:
        pass

    @property
    def inverse(self: Self) -> Self:
        """
        A new transformation that is the inverse of this transformation.
        """
        return self.__invert__()

    @abc.abstractmethod
    def __matmul__(self, other: AbstractTransform) -> AbstractTransform:
        """
        Compose multiple transformations into a single transformation.

        Parameters
        ----------
        other
            another transformation to compose with this one

        Examples
        --------

        Compose two transformations

        .. jupyter-execute::

            import astropy.units as u
            import named_arrays as na
            import optika

            t1 = optika.transforms.Translation(na.Cartesian3dVectorArray(x=5) * u.mm)
            t2 = optika.transforms.RotationZ(50 * u.deg)

            t_composed = t1 @ t2
            t_composed

        use the composed transformation to transform a vector

        .. jupyter-execute::

            v = na.Cartesian3dVectorArray(1, 2, 3) * u.mm

            t_composed(v)

        transform the same vector by applying each transformation separately
        and note that it's the same result as using the composed transformation

        .. jupyter-execute::

            t1(t2(v))
        """
        if isinstance(self, AbstractTransform):
            if isinstance(self, TransformList):
                self_normalized = self
            else:
                self_normalized = TransformList([self])
        else:
            return NotImplemented

        if isinstance(other, AbstractTransform):
            if isinstance(other, TransformList):
                other_normalized = other
            else:
                other_normalized = TransformList([other])
        else:
            return NotImplemented

        return self_normalized + other_normalized


@dataclasses.dataclass
class Transform(
    AbstractTransform,
):
    matrix: None | na.AbstractCartesian3dMatrixArray = None
    vector: None | na.AbstractCartesian3dVectorArray = None


@dataclasses.dataclass(eq=False, repr=False)
class Translation(AbstractTransform):
    """
    A translation-only vector transformation.

    Examples
    --------

    Translate a vector using a single transformation

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        displacement = na.Cartesian3dVectorArray(
            x=12 * u.mm,
            y=12 * u.mm,
        )

        transform = optika.transforms.Translation(displacement)

        square = na.Cartesian3dVectorArray(
            x=na.ScalarArray([-10, 10, 10, -10, -10] * u.mm, axes="vertex"),
            y=na.ScalarArray([-10, -10, 10, 10, -10] * u.mm, axes="vertex"),
        )

        square_transformed = transform(square)

        with astropy.visualization.quantity_support():
            plt.figure();
            plt.gca().set_aspect("equal");
            na.plt.plot(square, label="original");
            na.plt.plot(square_transformed, label="translated");
            plt.legend();

    |

    Translate a vector using an array of transformations

    .. jupyter-execute::

        displacement_2 = na.Cartesian3dVectorArray(
            x=na.ScalarArray([12, -12] * u.mm, axes="transform"),
            y=9 * u.mm,
        )

        transform_2 = optika.transforms.Translation(displacement_2)

        square_transformed_2 = transform_2(square)

        with astropy.visualization.quantity_support():
            plt.figure();
            plt.gca().set_aspect("equal");
            na.plt.plot(square, label="original");
            na.plt.plot(square_transformed_2, axis="vertex", label="translated");
            plt.legend();
    """

    displacement: na.Cartesian3dVectorArray = dataclasses.field(
        default_factory=na.Cartesian3dVectorArray
    )
    """The magnitude of the translation along each axis"""

    @property
    def explicit(self) -> Transform:
        return Transform(
            vector=self.displacement,
        )

    def __invert__(self: Self) -> Self:
        return type(self)(displacement=-self.displacement)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractRotation(AbstractTransform):
    """
    Any arbitrary rotation
    """

    angle: na.ScalarLike
    """The angle of the rotation"""

    def __invert__(self: Self) -> Self:
        return type(self)(angle=-self.angle)


@dataclasses.dataclass(eq=False, repr=False)
class RotationX(AbstractRotation):
    """
    A rotation about the :math:`x` axis
    """

    @property
    def explicit(self) -> Transform:
        return Transform(
            matrix=na.Cartesian3dXRotationMatrixArray(self.angle),
        )


@dataclasses.dataclass(eq=False, repr=False)
class RotationY(AbstractRotation):
    """
    A rotation about the :math:`y` axis
    """

    @property
    def explicit(self) -> Transform:
        return Transform(
            matrix=na.Cartesian3dYRotationMatrixArray(self.angle),
        )


@dataclasses.dataclass(eq=False, repr=False)
class RotationZ(AbstractRotation):
    """
    A rotation about the :math:`z` axis
    """

    @property
    def explicit(self) -> Transform:
        return Transform(
            matrix=na.Cartesian3dZRotationMatrixArray(self.angle),
        )


@dataclasses.dataclass(eq=False, repr=False)
class TransformList(
    AbstractTransform,
    optika.mixins.DataclassList[AbstractTransform],
):
    """
    A sequence of transformations
    """

    intrinsic: bool = True

    @property
    def extrinsic(self) -> bool:
        return not self.intrinsic

    @property
    def transforms(self) -> Iterator[AbstractTransform]:
        if self.intrinsic:
            return reversed(list(self))
        else:
            return iter(self)

    @property
    def explicit(self) -> Transform:
        matrix = na.Cartesian3dIdentityMatrixArray()
        vector = na.Cartesian3dVectorArray()

        for transform in reversed(list(self.transforms)):
            if transform is not None:
                transform = transform.explicit
                if transform.matrix is not None:
                    matrix = matrix @ transform.matrix
                if transform.vector is not None:
                    vector = matrix @ transform.vector + vector

        return Transform(
            matrix=matrix,
            vector=vector,
        )

    def __invert__(self: Self) -> Self:
        other = copy.copy(self)
        other.data = []
        for transform in self:
            if transform is not None:
                transform = transform.__invert__()
            other.append(transform)
        other.reverse()
        return other


@dataclasses.dataclass(eq=False, repr=False)
class Transformable(abc.ABC):
    @property
    @abc.abstractmethod
    def transform(self) -> None | AbstractTransform:
        """
        the coordinate transformation between the global coordinate system
        and this object's local coordinate system
        """



from __future__ import annotations
import abc
import dataclasses
import numpy as np
import astropy.units as u
import named_arrays as na
import optika.mixins

__all__ = [
    "AbstractMaterial",
    "Vacuum",
    "Mirror",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractMaterial(
    optika.mixins.Printable,
    optika.transforms.Transformable,
):
    @abc.abstractmethod
    def index_refraction(
        self,
        rays: optika.rays.AbstractRayVectorArray,
    ) -> na.ScalarLike:
        """
        the index of refraction of this material for the given input rays

        Parameters
        ----------
        rays
            input rays used to evaluate the index of refraction
        """

    @abc.abstractmethod
    def attenuation(
        self,
        rays: optika.rays.AbstractRayVectorArray,
    ) -> na.ScalarLike:
        """
        the attenuation coefficient of the given rays

        Parameters
        ----------
        rays
            input rays to calculate the attenuation coefficient for
        """

    @abc.abstractmethod
    def transmissivity(
        self,
        rays: optika.rays.AbstractRayVectorArray,
    ) -> na.ScalarLike:
        """
        The fraction of light that passes through the interface.

        Parameters
        ----------
        rays
            the input rays to calculate the transmissivity for
        """

    def refract_rays(
        self,
        rays: optika.rays.AbstractRayVectorArray,
        sag: optika.sags.AbstractSag,
        rulings: None | optika.rulings.AbstractRulings,
    ) -> optika.rays.RayVectorArray:
        rays_input = rays

        rays = sag.intercept(rays)

        distance = (rays.position - rays_input.position).length
        depth = rays.attenuation * distance
        rays.intensity = rays.intensity * depth * self.transmissivity(rays)

        n2 = self.index_refraction(rays)
        k2 = self.attenuation(rays)

        if rulings is not None:
            rays = rulings.rays_apparent(rays, index_refraction=n2)

        a = rays.direction
        n1 = rays.index_refraction
        normal = sag.normal(rays.position)

        r = n1 / n2
        c = -a @ normal
        b = r * a + (r * c - np.sqrt(1 - np.square(r) * (1 - np.square(c)))) * normal

        rays.direction = b / b.length
        rays.index_refraction = n2
        rays.attenuation = k2

        return rays


@dataclasses.dataclass(eq=False, repr=False)
class Vacuum(
    AbstractMaterial,
):
    @property
    def transform(self) -> None:
        return None

    def index_refraction(
        self,
        rays: optika.rays.RayVectorArray,
    ) -> na.ScalarLike:
        return 1

    def attenuation(
        self,
        rays: optika.rays.RayVectorArray,
    ) -> na.ScalarLike:
        return 0 / u.mm

    def transmissivity(
        self,
        rays: optika.rays.RayVectorArray,
    ) -> na.ScalarLike:
        return 1


@dataclasses.dataclass(eq=False, repr=False)
class Mirror(
    AbstractMaterial,
):
    @property
    def transform(self) -> None:
        return None

    def index_refraction(
        self,
        rays: optika.rays.RayVectorArray,
    ) -> na.ScalarLike:
        return -rays.index_refraction

    def attenuation(
        self,
        rays: optika.rays.RayVectorArray,
    ) -> na.ScalarLike:
        return rays.attenuation

    def transmissivity(
        self,
        rays: optika.rays.RayVectorArray,
    ) -> na.ScalarLike:
        return 1



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
import optika._tests.test_transforms


class AbstractTestAbstractRulings(
    optika._tests.test_transforms.AbstractTestTransformable,
):
    def test_diffraction_order(self, a: optika.rulings.AbstractRulings):
        assert na.get_dtype(a.diffraction_order) == int

    def test_ruling_normal(self, a: optika.rulings.AbstractRulings):
        position = na.Cartesian3dVectorArray() * u.mm
        result = a.ruling_normal(position)
        assert isinstance(result, na.AbstractCartesian3dVectorArray)
        assert np.all(result.length == 1)

    @pytest.mark.parametrize(
        argnames="rays",
        argvalues=[
            optika.rays.RayVectorArray(
                wavelength=500 * u.nm,
                position=na.Cartesian3dVectorArray(
                    x=na.linspace(-5, 5, axis="x", num=3) * u.mm,
                    y=na.linspace(-5, 5, axis="y", num=4) * u.mm,
                    z=0 * u.mm,
                ),
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            )
        ],
    )
    def test_rays_apparent(
        self,
        a: optika.rulings.AbstractRulings,
        rays: optika.rays.RayVectorArray,
    ):
        result = a.rays_apparent(rays, index_refraction=1)
        assert isinstance(result, optika.rays.AbstractRayVectorArray)
        assert np.all(result.wavelength == rays.wavelength)
        assert np.all(result.position == rays.position)
        assert np.any(result.direction.z != rays.direction.z)
        assert np.allclose(result.direction.length, 1)
        assert np.all(result.intensity == rays.intensity)
        assert np.any(result.index_refraction != rays.index_refraction)


class AbstractTestAbstractConstantDensityRulings(
    AbstractTestAbstractRulings,
):
    def test_ruling_density(
        self,
        a: optika.rulings.AbstractConstantDensityRulings,
    ):
        assert a.ruling_density.unit.is_equivalent(1 / u.mm)

    def test_ruling_spacing(
        self,
        a: optika.rulings.AbstractConstantDensityRulings,
    ):
        assert a.ruling_spacing.unit.is_equivalent(u.mm)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.rulings.ConstantDensityRulings(
            ruling_density=5000 / u.mm,
            diffraction_order=1,
        ),
        optika.rulings.ConstantDensityRulings(
            ruling_density=na.linspace(1, 5, axis="rulings", num=4) / u.mm,
            diffraction_order=na.ScalarArray(np.array([-1, 0, 1]), axes="m"),
        ),
    ],
)
class TestConstantDensityRulings(
    AbstractTestAbstractConstantDensityRulings,
):
    pass



import abc
import dataclasses
import astropy.units as u
import named_arrays as na
import optika
import optika.mixins

__all__ = [
    "AbstractRulings",
    "AbstractConstantDensityRulings",
    "ConstantDensityRulings",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractRulings(
    optika.transforms.Transformable,
):
    """
    Interface for the interaction of a ruled surface with incident light
    """

    @property
    @abc.abstractmethod
    def diffraction_order(self) -> na.ScalarLike:
        """
        the diffraction order to simulate
        """

    @abc.abstractmethod
    def ruling_normal(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractCartesian3dVectorArray:
        """
        vector normal to the plane of the rulings

        Parameters
        ----------
        position
            location to evaluate the normal vector
        """

    @abc.abstractmethod
    def rays_apparent(
        self,
        rays: optika.rays.AbstractRayVectorArray,
        index_refraction: float,
    ) -> optika.rays.RayVectorArray:
        """
        the apparent input rays from the given actual input rays

        Parameters
        ----------
        rays
            actual input rays
        index_refraction
            index of refraction of the output rays
        """


@dataclasses.dataclass(eq=False, repr=False)
class AbstractConstantDensityRulings(
    AbstractRulings,
):
    @property
    @abc.abstractmethod
    def ruling_density(self) -> na.ScalarLike:
        """
        the frequency of the ruling pattern
        """

    @property
    def ruling_spacing(self) -> na.ScalarLike:
        """
        the distance between successive rulings
        """
        return 1 / self.ruling_density

    def ruling_normal(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractCartesian3dVectorArray:
        """
        unit vector normal to the planes of the rulings

        Parameters
        ----------
        position
            the location to evalulate the ruling normal
        Returns
        -------

        """
        return na.Cartesian3dVectorArray(1, 0, 0)
        # ruling_density = self.ruling_density
        # return na.Cartesian3dVectorArray(
        #     x=ruling_density,
        #     y=0 * ruling_density.unit,
        #     z=0 * ruling_density.unit,
        # )

    def rays_apparent(
        self,
        rays: optika.rays.RayVectorArray,
        index_refraction: float,
    ) -> optika.rays.RayVectorArray:
        """
        apparent direction of incoming rays if there was no diffraction

        Parameters
        ----------
        rays
            incoming rays that will be diffracted by the rulings
        index_refraction
            the index of refraction after the rulings
        """
        a = rays.index_refraction * rays.direction
        diffraction_order = self.diffraction_order
        ruling_density = self.ruling_density
        ruling_normal = ruling_density * self.ruling_normal(rays.position)
        a = a + index_refraction * diffraction_order * rays.wavelength * ruling_normal
        length_a = a.length
        return optika.rays.RayVectorArray(
            wavelength=rays.wavelength,
            position=rays.position,
            direction=a / length_a,
            intensity=rays.intensity,
            index_refraction=length_a,
        )


@dataclasses.dataclass(eq=False, repr=False)
class ConstantDensityRulings(
    AbstractConstantDensityRulings,
):
    ruling_density: na.ScalarLike = 0 / u.mm
    diffraction_order: na.ScalarLike = 1
    transform: None | optika.transforms.AbstractTransform = None



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
import optika._tests.test_transforms
import optika.rays._tests.test_ray_vectors


class AbstractTestAbstractMaterial(
    optika._tests.test_transforms.AbstractTestTransformable,
):
    @pytest.mark.parametrize("rays", optika.rays._tests.test_ray_vectors.rays)
    class TestRayDependentMethods:
        def test_index_refraction(
            self,
            a: optika.materials.AbstractMaterial,
            rays: optika.rays.AbstractRayVectorArray,
        ):
            result = a.index_refraction(rays)
            assert isinstance(na.as_named_array(result), na.AbstractScalar)
            assert na.unit_normalized(result).is_equivalent(u.dimensionless_unscaled)

        def test_attenuation(
            self,
            a: optika.materials.AbstractMaterial,
            rays: optika.rays.AbstractRayVectorArray,
        ):
            result = a.attenuation(rays)
            assert isinstance(na.as_named_array(result), na.AbstractScalar)
            assert na.unit_normalized(result).is_equivalent(1 / u.mm)

        def test_transmissivity(
            self,
            a: optika.materials.AbstractMaterial,
            rays: optika.rays.AbstractRayVectorArray,
        ):
            result = a.transmissivity(rays)
            assert isinstance(na.as_named_array(result), na.AbstractScalar)
            assert na.unit_normalized(result).is_equivalent(u.dimensionless_unscaled)

        @pytest.mark.parametrize(
            argnames="sag",
            argvalues=[
                optika.sags.NoSag(),
            ],
        )
        @pytest.mark.parametrize(
            argnames="rulings",
            argvalues=[
                None,
            ],
        )
        def test_refract_rays(
            self,
            a: optika.materials.AbstractMaterial,
            rays: optika.rays.AbstractRayVectorArray,
            sag: optika.sags.AbstractSag,
            rulings: None | optika.rulings.AbstractRulings,
        ):
            result = a.refract_rays(
                rays=rays,
                sag=sag,
                rulings=rulings,
            )
            assert isinstance(result, optika.rays.AbstractRayVectorArray)
            assert np.all(result.index_refraction == a.index_refraction(rays))
            assert np.all(result.attenuation == a.attenuation(rays))


@pytest.mark.parametrize("a", [optika.materials.Vacuum()])
class TestVacuum(
    AbstractTestAbstractMaterial,
):
    pass


@pytest.mark.parametrize("a", [optika.materials.Mirror()])
class TestMirror(
    AbstractTestAbstractMaterial,
):
    pass



import abc
import dataclasses
import numpy.typing as npt
import matplotlib.axes
import matplotlib.lines
import named_arrays as na
import optika.transforms

__all__ = [
    "Plottable",
]


@dataclasses.dataclass(eq=False, repr=False)
class Plottable(abc.ABC):
    @property
    @abc.abstractmethod
    def kwargs_plot(self) -> None | dict:
        """
        Extra keyword arguments that will be used in the call to
        :func:`named_arrays.plt.plot` within the :meth:`plot` method.
        """

    @abc.abstractmethod
    def plot(
        self,
        ax: None | matplotlib.axes.Axes | na.ScalarArray[npt.NDArray] = None,
        transform: None | optika.transforms.AbstractTransform = None,
        component_map: None | dict[str, str] = None,
        **kwargs,
    ) -> na.AbstractScalar:
        """
        Plot the selected components onto the given axes.

        Parameters
        ----------
        ax
            The matplotlib axes to plot onto
        transform
            Any extra transformations to apply to the coordinate system before
            plotting
        component_map
            An optional mapping that relates the coordinate system of the
            :class:`Plottable` instance to the coordinate system of the
            :class:`matplotlib.axes.Axes` instance.
            If :obj:`None`, it is equivalent to ``dict(x="x", y="y", z="z")``.
        kwargs
            Additional keyword arguments that will be passed along to
            :func:`named_arrays.plt.plot()`
        """



import pytest
import abc
import numpy as np
import matplotlib.axes
import matplotlib.artist
import matplotlib.pyplot as plt
import astropy.units as u
import named_arrays as na
import optika.plotting


kwargs_plot_parameterization = [
    dict(),
    dict(color="red"),
]


class AbstractTestPlottable(abc.ABC):
    @pytest.mark.parametrize(
        argnames="ax",
        argvalues=[
            None,
            plt.subplots()[1],
        ],
    )
    @pytest.mark.parametrize(
        argnames="transform",
        argvalues=[
            None,
            optika.transforms.TransformList(
                [
                    optika.transforms.Translation(
                        na.Cartesian3dVectorArray(x=5 * u.mm),
                    ),
                    optika.transforms.RotationX(54 * u.deg),
                ]
            ),
        ],
    )
    @pytest.mark.parametrize(
        argnames="component_map",
        argvalues=[
            None,
            dict(x="y", y="x"),
        ],
    )
    class TestPlot(abc.ABC):
        def test_plot(
            self,
            a: optika.plotting.Plottable,
            ax: None | matplotlib.axes.Axes | na.ScalarArray,
            transform: None | optika.transforms.AbstractTransform,
            component_map: None | dict[str, str],
        ):
            result = a.plot(
                ax=ax,
                transform=transform,
                component_map=component_map,
            )

            if ax is None or ax is np._NoValue:
                ax_normalized = plt.gca()
            else:
                ax_normalized = ax
            ax_normalized = na.as_named_array(ax_normalized)

            for index in ax_normalized.ndindex():
                assert ax_normalized[index].ndarray.has_data()

            assert isinstance(result, na.AbstractScalar)
            assert result.dtype == matplotlib.artist.Artist
